/*
** SDMMC.c SD card interface 
** 
** 7/20/06 v1.4 LDJ 
** 7/14/07 v2.0 LDJ 
**
** AND20090129 adapted for usage of Explorer16 with
**             PICTail Daughter Board for SD/MMC (AC164122)
**             tested with PIM PIC32MX460F512L via SPI1
** AND20090129 gpStuff.h introduced for flexible frequency
** 
*/
#include <p32xxxx.h>
#include <gpStuff.h>
#include <explore.h>
#include <sdmmc.h>

/*
 * Pin-Assignments according to DS-51583b, DS-51775a
 * GP:  "general purpose chip" -> PIC32MX360F512L
 * USB: "USB On-The-Go"        -> PIC32MX460F512L
 *  log. GP-SPI1/2 USB-SPI1/2
 *  SCK   RF6/RG6  RD10/RG6
 *  SDI   RF7/RG7   RC4/RG7
 *  SDO   RF8/RG8   RD0/RG8
 *  ~CS   RB1/RB9   RB1/RB9
 *  CD    RF0/RG0   RF0/ -
 *  WD    RF1/RG1   RF1/ -
 */

// I/O definitions (for PIC32MX460F512L !!!)
#define SDWP        _RF1      // Write Protect input
#define SDCD        _RF0      // Card Detect input
#define SDCS        _RB1      // Card Select output
//+
#define TRIS_SCK    _TRISD10  // SCK pin direction
#define TRIS_SDI    _TRISC4   // SDI pin direction
#define TRIS_SDO    _TRISD0   // SDO pin direction
#define TRIS_CS     _TRISB1   // ~CS pin direction
#define TRIS_CD     _TRISF0   //  CD pin direction
#define TRIS_WD     _TRISF1   //  WD pin direction
//+
#define PCFG        (1)       // port configuration : ADC (Port B) digital bits


// SD card commands
#define RESET           0 // a.k.a. GO_IDLE (CMD0)
#define INIT            1 // a.k.a. SEND_OP_COND (CMD1)
#define READ_SINGLE     17
#define WRITE_SINGLE    24

// additional commands (not used)
#define SEND_CSD        9
#define SEND_CID        10
#define SET_BLEN        16
#define APP_CMD         55
#define SEND_APP_OP     41

// SD card responses
#define DATA_START      0xFE
#define DATA_ACCEPT     0x05

// timeouts
#define I_TIMEOUT       10000   
#define R_TIMEOUT       25000
#define W_TIMEOUT       250000

/*
 *
 */
void initSD( void)
{
    // PIC32MX460F512L !!!
    TRIS_SDO = 0;       // SDO output pin
    TRIS_SDI = 1;       // SDI input pin
    TRIS_SCK = 0;       // SCK output pin

    // Card I/O's init
    TRIS_CD = 1;        // Card Detect input
    TRIS_WD = 1;        // Write Protect input
    
    _PCFG1  = PCFG;     // make Card select as digital pin
    TRIS_CS = 0;        // make Card select an output pin
    
    SDCS = 1;           // disable SD card (AND20090129)
       
    // init the spi module for a slow (safe) clock speed first
    SPI1CON  = 0x8120; /*
      MSTEN=1//Mastermode; CKP=0; SSEN=0; 
      CKE=1//???; SMP=1//sample_middle???; MODE=8bit;
      DISSDO=0//???; SIDL=0; FRZ=0; ON=1;
    */
    SPI1BRG = FPB / 250000L / 2 - 1; // fSCK = fPB/(2*(1+SPI1BRG)): we want 250 kHz
}   // initSD


/*
 * send one byte of data and receive one back at the same time
 */
unsigned char writeSPI( unsigned char b)
{
    SPI1BUF = b;                  // write to buffer for TX
    while( !SPI1STATbits.SPIRBF); // wait transfer complete
    return SPI1BUF;               // read the received value
}// writeSPI

#define readSPI()   writeSPI( 0xFF)
#define clockSPI()  writeSPI( 0xFF)
#define disableSD() SDCS = 1; clockSPI()
#define enableSD()  SDCS = 0


/*
 *  c    command code
 *  a    byte address of data block
 */
int sendSDCmd( unsigned char c, LBA a)
{
    int i, r;

    // enable SD card
    enableSD();

    // send a comand packet (6 bytes)
    writeSPI( c | 0x40);    // send command 
    writeSPI( a>>24);       // msb of the address
    writeSPI( a>>16);       
    writeSPI( a>>8);
    writeSPI( a);           // lsb
    
    writeSPI( 0x95);        // send CMD0 CRC 

    // now wait for a response, allow for up to 8 bytes delay
    for( i=0; i<8; i++) 
    {
        r = readSPI();      
        if ( r != 0xFF) 
            break;
    }
    return ( r);         

/* return response
    FF - timeout 
    00 - command accepted
    01 - command received, card in idle state after RESET

other codes:
    bit 0 = Idle state
    bit 1 = Erase Reset
    bit 2 = Illegal command
    bit 3 = Communication CRC error
    bit 4 = Erase sequence error
    bit 5 = Address error
    bit 6 = Parameter error
    bit 7 = Always 0
*/
    // NOTE CSCD is still low!
} // sendSDCmd


/*
 * returns 0 if successful
 *         E_COMMAND_ACK   failed to acknowledge reset command
 *         E_INIT_TIMEOUT  failed to initialize
 */
int initMedia( void)
{
    int i, r;

    // 1. with the card NOT selected     
    disableSD(); 

    // 2. send 80 clock cycles start up
    for ( i=0; i<10; i++)
        clockSPI();

    // 3. now select the card
    enableSD();

    // 4. send a single RESET command
    r = sendSDCmd( RESET, 0); 
    disableSD();
    if ( r != 1)                // must return Idle
        return E_COMMAND_ACK;   // comand rejected

    // 5. send repeatedly INIT until Idle terminates
    for (i = 0; i < I_TIMEOUT; i++) 
    {
        r = sendSDCmd( INIT, 0); 
        disableSD();
        if ( !r) 
            break; 
    } 
    if ( i >= I_TIMEOUT)   
        return E_INIT_TIMEOUT;  // init timed out 

    // 6. increase speed 
    SPI1CON = 0;                // disable the SPI1 module
    SPI1BRG = FPB / 10000000L / 2 - 1;  // fSCK = fPB/(2*(1+SPI1BRG)): we want 10 MHz
    SPI1CON = 0x8120;           // re-enable the SPI1 module
    
    return 0;           
} // init media


/*
 *  a        LBA of sector requested
 *  p        pointer to sector buffer
 *  returns  TRUE if successful
 */
int readSECTOR( LBA a, char *p)
{
    int r, i;
    
#ifdef READ_LED    
    READ_LED = 0;
#endif
        
    // 1. send READ command
    r = sendSDCmd( READ_SINGLE, ( a << 9));
    if ( r == 0)    // check if command was accepted
    {  
        // 2. wait for a response
        for( i=0; i<R_TIMEOUT; i++)
        {
            r = readSPI();     
            if ( r == DATA_START) 
                break;
        } 

        // 3. if it did not timeout, read 512 byte of data
        if ( i != R_TIMEOUT)
        {
            i = 512;
            do{ 
                *p++ = readSPI();
            } while (--i>0);

            // 4. ignore CRC
            readSPI();
            readSPI();

        } // data arrived

    } // command accepted

    // 5. remember to disable the card
    disableSD();

#ifdef READ_LED // AND20090128 was  #ifdef READLED
    READ_LED = 1;
#endif

    return ( r == DATA_START);    // return TRUE if successful
} // readSECTOR



/*
 *  a        LBA of sector requested
 *  p        pointer to sector buffer
 *  returns  TRUE if successful
 */
int writeSECTOR( LBA a, char *p)
{
    unsigned r, i;

    // 0. check Write Protect
    if ( getWP())
        return FAIL;

    // 1. send WRITE command
    r = sendSDCmd( WRITE_SINGLE, ( a << 9));
    if ( r == 0)    // check if command was accepted
    {  
        // 2. send data
        writeSPI( DATA_START);
        
        // send 512 bytes of data
        for( i=0; i<512; i++)
            writeSPI( *p++);

        // 3. send dummy CRC
        clockSPI();
        clockSPI();
    
        // 4. check if data accepted
        r = readSPI(); 
        if ( (r & 0xf) == DATA_ACCEPT)
        {   
#ifdef WRITE_LED    
            WRITE_LED = 0;
#endif
            // 5. wait for write completion
            for( i=0; i<W_TIMEOUT; i++)
            { 
                r = readSPI();
                if ( r != 0 )
                    break;
            } 
#ifdef WRITE_LED
            WRITE_LED = 1;
#endif
        } // accepted
        else
            r = FAIL;

    } // command accepted

    // 6. disable the card
    disableSD();

    return ( r);      // return TRUE if successful

} // writeSECTOR


// SD card connector presence detection switch
// returns  TRUE card present
//          FALSE card not present
int getCD( void) 
{
    return !SDCD;
}


// card Write Protect tab detection switch
// returns  TRUE write protect tab on LOCK
//          FALSE write protection tab OPEN
int getWP( void)
{
    return SDWP;
}
